;calculates a 16-bit signature.
;Works for my TERSE v3.1 for the '86 processor.
;Note that the algorithm used in this version of TERSE for
;calculating the signature, could be improved.
;see notes below.
;To assemble:	MASM SIGCAL16;
;to link:	LINK SIGCAL16;

.MODEL SMALL
.STACK
.DATA
	DB   10 DUP(0)
asciitbl DB  7 DUP(0)
	DB   "$","$"
intromsg DB 0Ah,0Dh
	 DB "Signature calculator for 16-bit '86 TERSE v3.1.",0Ah,0Dh
	 DB "Type-in active-set of messages, crtl-z to clear, crtl-x to quit"
	 DB 0Ah,0Dh
intr2    DB "Do not enter remote-output, nor Notification messages.",0Ah,0Dh
intro3	 DB "Node range = 0 - 31. Terminal range = 1 - 8."
	 DB 0Ah,0Dh
	 DB "Processor range 1 - 63",0Dh,0Ah
	 DB "Enter messages in order of oldest to latest.",0Dh,0Ah
	 DB "-- i.e. they are in the buffer in this order, and",0Dh,0Ah
	 DB "   the signature is calculated in this order."
	 DB 0Ah,0Dh,0Ah,0Dh,"$"
destnode DB "Destination-node: $"
destterm DB " Destination-terminal: $"
sigtxt   DB "  SIGNATURE:$"
newline  DB 0Dh,0Ah,"$"

signature DW 0          ;16-bit signature
input     DB 0          ;binary input
header    DW 0         ;used for sig-calc
termhi    DB 0      ;bit-2 of term

msg2 DB 0       ;header-low  17/12/96
msg3 DB 0	;header-high 17/12/96

szproc DB "Processor number: $"

.CODE
start:
	mov ax,@DATA
	mov ds,ax
start2:
;intro message...
	mov ah,9
	lea dx,intromsg
	int 21h

;ask for processor number. v3.1 '86 TERSE computes signature using
;the entire header of the message, being msg2 and msg3 (see TERS8631.ASM)
;however, msg3 contains the processor number, which really doesn't
;matter as we have restricted sig. calc to local messages only
;i.e. it will always be the same value.
;however, status quo for now, is have to ask for it...
	mov  ah,9
	lea  dx,szproc
	int  21h
	call getinput		;-->input (binary value)
	mov  al,input
	and  al,00111111b	;PU is in bits 0--5.
	mov  msg3,al

	mov  ah,9
	lea  dx,newline
	int  21h

start3:
;Ask for destination node....
	mov ah,9
	lea dx,destnode
	int 21h
	call getinput           ;comes back with "input", in binary...
	mov al,input
	and  ax,00011111b       ;node # is in bits 0--4. 17/12/96
	mov  header,ax
	mov  msg2,al            ;17/12/96

;ask for destination terminal...
	mov ah,9
	lea dx,destterm
	int 21h
	call getinput
	mov  al,input
	  cmp al,0	;17/12/96 must be 1--8
	  je  getout	;/
	  dec al	;17/12/96 as stored 0--7.
	and  ax,0111b
	  clc
	  rcr al,1
	  rcr al,1
	  rcr al,1

	  rcr al,1      ;want terminal# in bits 5--7.  17/12/96

;17/12/96          ;bit-2 of term is now in carry...
;/          mov termhi,0
;/          rcl termhi,1
	or   header,ax

	or  msg2,al   ;combine terminal# and node#.



;calculate the signature....
	mov ax,signature
	test ax,4000h	;8000h
		;mathematically, should be 8000h, however originally
		;I used 4000h and for now stuck with it.

	jz   sig1
	xor  ax,0010010000010001b
sig1:	xor ax,WORD PTR msg2	;xor	ax,header
		;this also should be improved.
		;no need to include msg3 in the calculation.
		;should just have xor al,msg2.

	;  rcr termhi,1 	     ;bring back bit6 of node-->carry
	;rcl  ax,1
	;17/12/96 do the shift further down....for looping back...
	mov  signature,ax      ;in bits 0--14.
	
;convert it to ascii....
	mov ax,signature
	mov dx,0
	call bin2dec
	mov  BYTE PTR asciitbl+7,"$"    ;hex o/p stuffs this up.

;display result...
	mov ah,9
	lea dx,sigtxt
	int 21h
	mov ah,9
	lea dx,asciitbl
	int 21h
	  ;also display signature in hex...
	  mov bx,signature
	  mov cl,4
	  rol bx,cl
	  mov ax,bx
	  and ax,000Fh
	    cmp al,9
	    jbe xx
	    add al,7
xx:       add al,30h
	  mov asciitbl+1,al
	  rol bx,cl
	  mov ax,bx
	  and ax,000Fh
	    cmp al,9
	    jbe yy
	    add al,7
yy:       add al,30h
	  mov asciitbl+2,al
	  rol bx,cl
	  mov ax,bx
	  and ax,000Fh
	    cmp al,9
	    jbe zz
	    add al,7
zz:       add al,30h
	  mov asciitbl+3,al
	  rol bx,cl
	  mov ax,bx
	  and ax,000Fh
	    cmp al,9
	    jbe mm
	    add al,7
mm:       add al,30h
	  mov asciitbl+4,al
	  mov BYTE PTR asciitbl+5,"h"
	  mov BYTE PTR asciitbl+6,0Ah
	  mov BYTE PTR asciitbl+7,0Dh
	  mov BYTE PTR asciitbl+8,"$"
	  mov ah,9
	  lea dx,asciitbl
	  int 21h

	shl signature,1 ;17/12/96

	jmp start3
getout:
	mov ax,4C00h
	int 21h

getinput:
	mov input,0
getinput2:
	mov ah,0
	int 16h
	cmp ax,2C1Ah    ;crtl-z
	jne notcrtlz
	mov signature,0 
	mov ah,9
	lea dx,newline
	int 21h
	pop ax          ;dump return address
	jmp start2
notcrtlz:
	cmp ax,2D18h    ;crtl-x
	jne notcrtlx
	mov ah,9
	lea dx,newline
	int 21h
	jmp getout
notcrtlx:
;test if outside range 0-9...
	cmp al,30h
	jb  below0
	cmp al,39h
	ja  above9
inrange:
;echo it to screen...
	  mov ah,0Eh
	  push ax
	  mov bx,02
	  int 10h
	  pop ax
	and al,0Fh      ;convert to binary
	 mov ah,input
	 and ah,0Fh    ;see if 2nd char already there
	 jnz second
	 ;mov cl,4
	 ;shl al,cl
	mov input,al    ;save in lo-nibble.
	jmp getinput2
second:
	mov cl,4
	shl input,cl
	mov ah,input
	 or al,ah      ;combine them
	mov input,al    ;save. ...value is in bcd format.
below0:                ;any char outside range, terminates entry
above9:                ;        /
;convert input to binary....
	mov al,input
	and al,0F0h
	mov cl,4
	ror al,cl
	mov dl,10
	mul dl          ;result --> ax
	mov bl,input
	and bl,0Fh
	add al,bl       ;now have binary value.
	mov input,al    ;       /

	ret    
;...................................................................
bin2dec PROC            ;requires binary number in dx:ax...
	lea di,asciitbl
	mov bx,di
	add di,6  ;8
	;mov BYTE PTR [di],0
	;dec di
ssss:
	mov cx,10
	div cx
	add dl,30h
sssss:
	mov [di],dl
	dec di
	mov dx,0
	cmp ax,0
	jne ssss
	mov dl," "
	cmp bx,di
	jbe sssss
	ret
bin2dec ENDP

	END  start
