;TERS8631.ASM, for x86 embedded. Written for MASM v6.1/TASM v3.1.
; VERSION 1.0, Oct. 1996
;Uses MESS scheduling.
;Now has timesliced multitasking of virtual TERSE processors.
;Works with small-C.

;Message structure:
;BIT:	    07 06 05 04 03 02 01 00
;ADDR n+3:  0  s  p  p	p  p  p  p     header high (msg3)
;ADDR n+2:  t  t  t  n	n  n  n  n     header low  (msg2)
;ADDR n+1:  d  d  d  d	d  d  d  d     data   high (msg1)
;ADDR n:    d  d  d  d	d  d  d  d     data   low  (msg0)

;Legend:						       value:
; s = 0 sync message, = 1 async message 		      (0 -- 1)
; p = 0 empty slot, =111111b remote i/p msg, else is dest.-pu (1--62)
; t = destination terminal, less 1.			      (0 -- 7)
; n = destination node.					      (0 -- 31)


;......EQUATES YOU MAY HAVE TO MODIFY..............
SPEEDUP     EQU 55	;how many times we are speeding-up int8.
CIRCBUFSIZE EQU 128	;size of circular buffer(one for each virt-PU).
STACKSIZE   EQU 512	;each virt-PU has its own stack also.
TIMESLICE   EQU 20	;#msec timeslice (for SPEEDUP=55).

;..............EQUATES USED BY TERSE..................................
SYNC	     EQU  0	;used in POSTMESSAGE.
ASYNC	     EQU  1	;/
FALSE	     EQU  0
TRUE	     EQU  1
;........................................................................

;..........MACROS USED BY TERSE..................................

;note, uses ax,bx,es...
POSTMESSAGE MACRO pum,nodem,termm,asyncm,datawordm
     LOCAL msg1,msg2
     IF ((pum LE 0) OR (pum GT 61))	  ;assemble-time limit testing.
	EXITM				  ;/
     ELSEIF ((nodem LT 0) OR (nodem GT 31))   ;/
	EXITM				  ;/
     ELSEIF ((termm LT 1) OR (termm GT 8))	  ;/
	EXITM				  ;/
     ELSEIF ((asyncm LT 0) OR (asyncm GT 1))  ;/
	EXITM				  ;/
     ELSEIF pum EQ THISPU			;true, then local msg.
	mov  bx,pcircout
	cmp  bx,OFFSET PBOTTOM	    ;tail-end of circ.buffer.
	jne  msg1
	mov  bx,OFFSET PTOP	;(note that PTOP is lowest address).
msg1:	mov  ax,datawordm
	mov  WORD PTR [bx],ax		;msg0,msg1
	inc  bx
	inc  bx
	mov  BYTE PTR [bx],(nodem + ((termm - 1) * 32))	;msg2
	inc  bx
	mov  BYTE PTR [bx],(pum + (asyncm*64))	;msg3
	inc  bx
	mov  pcircout,bx		;new tail-end.
     ELSE			;false, so remote o/p msg.
msg1:
	mov  ax,SEG dataISR		;get ds of destination PU.
	mov  es,ax			;/
	mov  bx,OFFSET es:dsPU		;/
	mov  ax,es:[bx + ((pum - 1)*2)]	;/   lookup pu-table.
	mov  es,ax			;/ -->es

;deliver msg directly to circular buffer of dest. PU...
	mov  si,es:pcircin	;/
	cmp  si,OFFSET PTOP		;/
	ja   msg2		;/
	mov  si,OFFSET PBOTTOM	;/
msg2:	sub  si,4		;/
	mov  ax,datawordm
	cli			;prevent a task-switch.
	mov  es:[si],ax 	;/ msg0,msg1
	mov  es:[si+2],BYTE PTR (nodem + ((termm - 1) * 32))	;msg2
	mov  es:[si+3],BYTE PTR ((pum + (asyncm*64)) OR 00111111b) ;msg3
	mov  es:pcircin,si	;update leading-edge of buffer.
	sti

;need to set flag in extmsgwtg. ie, msg to node1 must set bit1...
;but only if it is sync msg...
	IF  asyncm EQ 0
	  mov  cx,nodem		;dest. node.
	  mov  ax,1		;/
	  shl  ax,cl		;convert to a flag.
	  or   es:extmsgwtg,ax	;must leave existing flags alone.
	  xor  cx,cx		;small-C needs inline asm exit with cx=0.
	ENDIF
     ENDIF
ENDM
;note that TERM is stored one-less. EX: TERM 1 is stored as 00.

NORMALEXIT MACRO		;use this one for normal operation.
	jmp  FAR PTR nodereturn
ENDM

REGSSTRUC STRUC 	;template for task block.
sax  DW  0
sbx  DW  0
scx  DW  0
sdx  DW  0
sbp  DW  0
ssi  DW  0
sdi  DW  0
ssp  DW  0
sss  DW  0
sip  DW  0
scs  DW  0
sds  DW  0
ses  DW  0
sflags DW 0
REGSSTRUC ENDS

initTASKS MACRO
;note that dataISR is current data segment when run this macro.
;initialise PU-table.... (only 2 PUs)
	mov  ax,SEG dataPU1
	mov  dsPU,ax
	mov  ax,SEG dataPU2
	mov  dsPU+2,ax
;initialise task-table.... (only 2 tasks here)
	mov  ax,SEG dataPU1  ;in DOS system initialise task table at
	mov  dstasks,ax 	;/ runtime.
	mov  ax,SEG dataPU2	;/ In ROM embedded system, maybe
	mov  dstasks+2,ax	;/ hardcode data seg. addresses.
	mov  ax,SEG codePU1
	mov  cstasks,ax
	mov  ax,SEG codePU2
	mov  cstasks+2,ax
;initialise task-blocks...
;firstly for PU1...
	mov  ax,SEG dataPU1
	mov  es,ax
	mov  es:regs.sds,ax
	mov  es:regs.sss,ax
	mov  bx,OFFSET stackPU + STACKSIZE - 2
	mov  es:regs.ssp,bx
	mov  es:regs.scs,cs		;start running inside TERSE.
	mov  es:regs.sip,OFFSET start1	;/
	;also, save codePU1 in PU1 data seg., for easy access...
	mov  ax,SEG codePU1	;(current-pu cs is also saved in
	mov  es:csthispu,ax	; dataISR seg.,in cstask, & all in cstasks)
	mov  es:pcircin,OFFSET PBOTTOM	;initialise circ.buff
	mov  es:pcircout,OFFSET PBOTTOM	;   /
;same for PU2...
	mov  ax,SEG dataPU2
	mov  es,ax
	mov  es:regs.sds,ax
	mov  es:regs.sss,ax
	mov  bx,OFFSET stackPU + STACKSIZE - 2
	mov  es:regs.ssp,bx
	mov  es:regs.scs,cs		 ;start running inside TERSE.
	mov  es:regs.sip,OFFSET start1	 ;/
	;also, save codePU2 in PU2 data seg., for easy access...
	mov  ax,SEG codePU2	;(current-pu cs is also saved in
	mov  es:csthispu,ax	; dataISR seg.,in cstask, & all in cstasks)
	mov  es:pcircin,OFFSET PBOTTOM	;initialise circ.buff
	mov  es:pcircout,OFFSET PBOTTOM	;   /
ENDM

;##################################################################
;############################# TERSE ##############################

;...............STACK SEGMENT FOR TERSE....................
;this is the startup stack, but each virt-PU has its own.
stackDEF  SEGMENT STACK
	DB 64 DUP(?)
stackDEF  ENDS

;.................DATA SEGMENT FOR TERSE TIMESLICER....................
;note that there is only one copy of TERSE, with its own data segment
;for the timeslicer interrupt-routine; dataISR.
;Each virtual processor has its own data segment also, that TERSE uses.
;note that all data SEGMENTs are private by default with MASM, and are
;not combined.

;....................................................................
dataISR SEGMENT	 'data'	;timeslicer data.
;timerisr uses these....
globaltime0 EQU 0	;future finer time granularity.
globaltime1 DB 0	;granularity = 1mSec.
globaltime2 DB 0	;granularity = 256mSec.(1/4Sec)
globaltime3 DB 0	;granularity = 64Sec (appr. 1minute)
timercnt    DW SPEEDUP	;how many times int8 is speed-up.
timeslicecnt DW TIMESLICE ;#msec between context switches.
int8off      DW	0	;timerisr old vector.
int8seg      DW	0	;/

dstask	     DW 0	;hold ds of current virt-pu.
cstask	     DW 0	;holds cs of current virt-pu.

;PU table...
NUMPUS	 EQU 2		   ;put ds of each pu in here.(done by initTASKS)
dsPU	 DW  NUMPUS DUP(0) ;/ (in order of PU #)

;task-table...
offsettasks DW 0		;which task running.0=pu1,2=pu2
dstasks  DW  0,0	;...as many tasks as required...(done by initTASKS)
cstasks  DW  0,0	;...must have same # of entries here.
NUMTASKS EQU (OFFSET cstasks - OFFSET dstasks) / 2
dataISR ENDS

;.................CODE SEGMENT FOR TERSE.............................
;TERSE itself is in its own code segment, codeTERSE, and each
;virt-PU is in its own code seg. They are not combined.

codeTERSE SEGMENT 'code'

start:				      ;inital entry point, only exec. once.
	cli			      ;don't use stackDEF. assume PU1
	mov  ax,SEG dataPU1	      ;/ is running initially.
	mov  ss,ax		      ;/
	mov  sp,OFFSET (stackPU + STACKSIZE - 2)  ;/
	sti			      ;/
	ASSUME ss:dataPU1	      ;/
	mov  ax,SEG dataISR	     ;in DOSless system, use RAM address.
	mov  ds,ax	     ;/
	ASSUME ds:dataISR,cs:codeTERSE,es:NOTHING
	mov  timercnt,SPEEDUP
	mov  timeslicecnt,TIMESLICE
	initTASKS	;segment values have
			;to be loaded into task-blocks and task-table
			;at runtime.
;&&
; jmp TEMPBYPASS ;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

;hook timer for TERSE timeslicer....
;NOTE, embedded system may have another (faster) timer available...
	;will use int-8, the PC 55mSec interrupt...
	mov  ax,0		;segment 0.
	mov  es,ax		;/
	mov  di,4*8		;ivt entry.
	mov  ax,es:[di] 	;get original. (NOTE we do not use
	mov  int8off,ax	;/		DOS services here,
	mov  ax,es:[di+2]	;/		as target system
	mov  int8seg,ax	;/		may not have DOS).
	cli
	mov  WORD PTR es:[di],OFFSET timerisr	;new vector.
	mov  es:[di+2],cs		;/
	sti			;allow interrupts.
	nop

;now, this is neat... let's speed-up the 55mSec timer to 1mSec...
	cli
	mov  al,36h		;select timer0, mode 3.
	out  43h,al		;/
	jmp  SHORT $+2		;delay.
	jmp  SHORT $+2		;/
	mov  ax,65532/SPEEDUP	;1192, gives 1mSec interrupts.
	out  40h,al		;/ (65532/55 = 1192)
	jmp  SHORT $+2		;/ (where 55 is how many times
	jmp  SHORT $+2		;/    we are speeding up above
	mov  al,ah		;/    the 55mSec rate)
	out  40h,al		;/
	sti
;*NOTE* USE THIS AT YOUR OWN RISK ON A PC. LOOK AT TIMERISR.

;&&
;TEMPBYPASS:

;assume this is PU1 running initially...
	mov  ax,SEG codePU1
	mov  cstask,ax		;in dataISR seg., holds current task cs.
	mov  ax,SEG dataPU1
	mov  dstask,ax		;in dataISR segment, holds current task ds.
	mov  ds,ax		;from now, access data seg. of pu1.
	jmp  start1
;note, all other tasks will commence from start1...

;......................TERSE TIMESLICER...............................

;TESTIT LABEL FAR  ;&&

timerisr:	;hooks int-8 on PC.
;int-8 has been speeded-up to 1mSec, so requires special handling here...
	push ds
	push ax
	  mov  ax,SEG dataISR		;get data segment addressability.
	  mov  ds,ax			;/
;tell the assembler what our data segment is...(nothing to do with runtime)
	  ASSUME ds:dataISR

; jmp doit ;&&

	    add	globaltime1,1		;note we're not using globaltime0.
	    adc WORD PTR globaltime2,0	;/
	    dec timeslicecnt
	    dec	timercnt
	    jz	tmr00
	      mov	al,20h		;/ process the interrupt here.
	      out	20h,al		;/
	      jmp  SHORT tmr01
tmr00:	    mov	timercnt,SPEEDUP	;reload.
	  pushf 			;as int8 routine has iret.
	  call DWORD PTR int8off	;call original vector.
tmr01:	cmp  timeslicecnt,0
	je   doit
	pop  ax
	pop  ds
	iret


;let's do a task-switch ...
doit:	mov  timeslicecnt,TIMESLICE	;reload.
	mov  ax,dstask		;get ds of current task.
	mov  ds,ax
;tell the assembler what our data segment is...(nothing to do with runtime)
  ASSUME ds:dataPU1
	mov  regs.sbx,bx		;save state of current task.
	mov  regs.scx,cx		;/    (the order here is
	mov  regs.sdx,dx		;/	     very important!)
	mov  regs.sbp,bp		;/
	mov  regs.ssi,si		;/
	mov  regs.sdi,di		;/
	mov  regs.sss,ss		;/
	mov  regs.ses,es		;/
	pop  regs.sax	;saved-ax	;/  these were pushed at entry.
	pop  regs.sds	;saved-ds	;/   /
	pop  regs.sip	;return-ip	;/
	pop  regs.scs	;return-cs	;/
	pop  regs.sflags ;return-flags	;/
	mov  regs.ssp,sp		;/

	  mov  ax,SEG dataISR		;isr data segment addressability.
	  mov  ds,ax			;/
;tell the assembler what our data segment is...(nothing to do with runtime)
  ASSUME ds:dataISR
	lea  bx,dstasks 		   ;get next entry in task table.
	add  offsettasks,2		   ;/
	cmp  offsettasks,(NUMTASKS*2)	 ;/
	jb	tmr2			   ;/
	mov  offsettasks,0		   ;/
tmr2:	add  bx,offsettasks		   ;/
	mov  ax,[bx + (NUMTASKS*2)]	;record ds/cs for new task.
	mov  cstask,ax			;/
	mov  ax,[bx]			;/
	mov  dstask,ax			;/

	mov  ds,ax			;get data addressability new task.
;tell the assembler what our data segment is...(nothing to do with runtime)
  ASSUME ds:dataPU1
	mov  bx,regs.sbx		;get state of next task.
	mov  cx,regs.scx		;/
	mov  dx,regs.sdx		;/
	mov  bp,regs.sbp		;/
	mov  si,regs.ssi		;/
	mov  di,regs.sdi		;/
	mov  sp,regs.ssp		;/
	mov  ss,regs.sss		;/
	mov  es,regs.ses		;/
	push regs.sflags		;get iret return.
	push regs.scs			;/
	push regs.sip			;/
	push regs.sax
	mov  ax,regs.sds
	mov  ds,ax
	pop  ax
	iret

;........................................................................
;*****MAIN LOOP*******
;tell the assembler what our data segment is...(nothing to do with runtime)
	ASSUME ds:dataPU1
start1:
				;main program entry point.


;....................................................................
nodereturn LABEL FAR		;<<<<<<<<<< Nodes return to here <<<<<
	mov  sp,OFFSET (stackPU + STACKSIZE - 2)	;remove any junk.

start2: 			;program entry point 2.
;*ASSUMPTION* here, that previous node hasn't changed ds!
;(you can reload it from dstask in dataISR).
	mov   dx,0		;initially, make signature = zero.
	mov   _fSIGFAIL,FALSE	;clr all error flags.
	mov   _fTIMEOUT,FALSE	;/
;make sure es set correctly...
	mov  ax,csthispu	;code seg. of current PU saved here.
	mov  es,ax		;/ we need to get at data in code seg.

;calculate the signature....
	mov  di,pcircin 	;leading-edge. remote i/p msgs added here.
;check limits...
ca1:	cmp  di,pcircout
	je   ca9		;finished
	cmp  di,OFFSET PBOTTOM
	jb   ca2
	mov  di,OFFSET PTOP
	jmp  ca1		;need to test for finished again.
ca2:
;search through buffer...
	mov  al,[di+3]		;msg3.
	cmp  al,0		;is it empty slot?
	jz   ca3		;  ignore.
	test al,80h		;is it a transform?		#40
	jnz  ca3		;  ignore.			#40
	test al,40h		;ignore async.msg (bit6 set).
	jnz  ca3		;/
	cmp  al,3Fh		;dest.pu (=3F if remote i/p).
	je   ca3		;/ ignore.

calcsig:
;entered with di-->msg0, dx=accumulated signature (starts =0)
	  test	dh,40h	 ;NOTE that this should really be 80h,
			 ;mathematically, but I used 40h originally
			 ;and have stuck with it. To change to 80h,
			 ;will require same change in SIGCAL16.ASM.
			 ;... and the sig.tables will have to be
			 ;changed.
	  jz	calc1
	  xor	dx,0010010000010001b	;tapping points.
calc1:	  xor	dx,[di+2]	;msg2,msg3.
	  shl	dx,1
	  ;ret		;note that lsb=0, sig is 15 bits, bits 1--15.

ca3:	add  di,4		;msg0 of next msg, going toward pcircout.
	jmp  ca1
ca9:	shr  dx,1	;make signature bits 0--14.


;have come out of above with a signature in dx, now lookup node...
lookup:
	xor   cx,cx
	mov   cl,es:NUMENTRIESVAR ;this is in codePUx.
	mov   si,OFFSET signatures
	push  ds	;save.
	push  es
	pop   ds	;lodsw reads ds:si.
look1:	lodsw				;start at beginning of array.
	xor   ax,dx			;compare signatures.
	jz    sigfnd			;signature found.
	loop  look1
	pop   ds		;restore.
;signature not found...
	mov   destnode,0
	mov   _fSIGFAIL,TRUE		;... go to node0.
	jmp   SHORT gonode
;.......................
sigfnd: pop   ds		;restore.
	dec  si 		;want si-->found sig.
	dec  si 		;/
;are we in a context where we need to check any ext.i/p arrivals?...
	xor  bx,bx
	mov  bl,es:NUMENTRIESVAR
	shl  bx,1		;	* 4
	shl  bx,1		;/
	mov  ax,es:[si+bx]	;get extmsg entry from sig.tabl
	mov  extmsgsaved,ax	;save it.
	or   ax,ax		;test if ax==0.
	jnz  ext1
;we are going to ignore all possible external i/ps...lookup nodenum...
	xor  bx,bx
	mov  bl,es:NUMENTRIESVAR
	shl  bx,1		; * 2
	mov  ax,es:[si+bx]			;get node.
;JUST IN CASE you set the WX flag in nodenum, it is not appropriate...
	and  ax,07FFFh		;"mistake proofing" (refer Figure 37)
	mov  destnode,ax
;.......................
gonode:
	jmp  getmsgs		;extract all msgs for this node.
retgetmsgs:			;...returns here.
go4:	mov  di,OFFSET nodeptr	;jump table is in code seg.,current pu.
	mov  bx,destnode	;
	shl  bx,1		;each entry in jump table 2 bytes.
	mov  cx,csthispu	;it was in es, but got overwritten.
	mov  es,cx		;/
	mov  ax,es:[di+bx]	;get entry in jump table.
	mov  ipnxtnode,ax	;form far address,  csthispu:ipnxtnode
	jmp  DWORD PTR ipnxtnode ;jump to next node to execute.>>>>>>>

;...................................................................




;........................
;extmsg entry in lookup table was non-0, so look for rem.i/p arrivals...
ext1:	and  ax,extmsgwtg	;compare with actual arrivals.
	jz   wtg1		;no appropriate ext.msgs arrived.
ext2:
	mov  bx,ax		;load destnode, clr bit in extmsgwtg.
	neg  ax 		;/ my clever code to extract 1 flag!
	and  ax,bx		;/  /		  ("rightmost" flag)
	;ax has the bit I need to clr in extmsgwtg...
	xor  extmsgwtg,ax	;/  /
	xor  bx,bx		;/ now load destnode.
ext3:	inc  bx			;/  /
	shr  ax,1		;/  /
	jnc  ext3		;/  /
	dec  bx			;/  /
	mov  destnode,bx	;/  /
;Note, flag of lowest-numbered node has highest priority always, as the
;above code extracts the rightmost set flag in extmsgwtg.
	jmp  gonode

;.........................
wtg1:		;some ext.i/ps are in-scope, but none have arrived...
	xor  bx,bx
	mov  ax,csthispu	;it was in es, but got overwritten.
	mov  es,ax		;/
	mov  bl,es:NUMENTRIESVAR
	shl  bx,1	       ;* 2
	mov  ax,es:[si+bx]		;get corresponding nodenum entry.
	mov  nodenumsaved,ax		;save it.
;If WX flag in nodenum =0, don't wait for rem.i/ps to arrive...
	test ax,8000h			;bit15 set if WX.
	jz   gonode
;.......
wtg2:		;we have to wait, but bits0--14 of nodenum are a timeout...
	  mov  ax,SEG dataISR
	  mov  es,ax
	mov  bl,es:globaltime0		;snapshot of globaltime.
	mov  bh,es:globaltime1		;/    (we need this for calc
	mov  WORD PTR localref0,bx	;/	    the timeout).
	mov  bx,WORD PTR es:globaltime2	;/
	mov  WORD PTR localref2,bx	;/
;......
wtg3:	;calc if we have timed out...
	;granularity of timeout can be adjusted....
	;let's use bits 14&13 as granularity field.
	mov  cx,nodenumsaved	;nodenum entry was saved this variable.
	and  cx,1FFFh		;mask off bits13-15.(15 is WAIT flag)
      ;if nodenum timeout ==0, wait forever...
	jz   wtg4
	and  ch,01100000b
	cmp  ch,0
	jne  wtg3b		;finest granularity ().
	  mov  bl,es:globaltime0	;this one not implemented.<<<<<
	  mov  bh,es:globaltime1
	  sub  bx,WORD PTR localref0
	  jmp  SHORT wtg3e
wtg3b:	cmp ch,00100000b	;2nd finest gran (1mSec).
	jne wtg3c
	  mov  bx,WORD PTR es:globaltime1  ;(timeout range 1mSec --
	  sub  bx,WORD PTR localref1	   ;		     8191mSec)
	  jmp  SHORT wtg3e
wtg3c:	cmp ch,01000000b
	jne wtg3d
	  mov  bx,WORD PTR es:globaltime2  ;(256mSec, about 1/4sec)
	  sub  bx,WORD PTR localref2	   ;(range 1/4sec -- 2048sec)
	  jmp  SHORT wtg3e
wtg3d:	  mov  bl,es:globaltime3  ;coarsest granularity (65.5sec,about 1min)
	  sub  bl,localref3		   ;(range 1min -- 256min)
	  xor  bh,bh
wtg3e:	cmp  ax,bx
	jbe  wtg4			;check arrival rem.i/p...again.
;waiting on ext.i/p, have timed out...
	mov  _fTIMEOUT,TRUE
	jmp  go4		;re-execute previous node!!!! <<<<
				;(i.e. don't update destnode)
	;<<<<THIS IS DIFFERENT FROM FIGURE 37 IN TEXT 1ST ED.<<<<<
;....
wtg4:	;check if any in-scope rem. i/ps have arrived...
;NOTE again, we deviate slightly from Figure 37 in text (1st ed)
;as have repeated some code here rather than branching back...
	mov  ax,extmsgsaved	;previously saved extmsg entry.
	and  ax,extmsgwtg	;compare with actual arrivals.
	jz   wtg3		;no appropriate ext.msgs arrived.
	;yes, a rem.i/p has arrived!!...
	jmp  ext2		;translate rem.i/p --> dest.node.

;....
;*****END MAIN LOOP*****
;.........................................................................

;.........................................................................

getmsgs:
;extract msgs for destnode, into msgterm"n" variables..
;removes any transforms encountered -- they occupy one complete slot.
;places async msgs on stack, count into msgterm"n" variable.

;can't be sure what es = on entry...
	mov  ax,csthispu
	mov  es,ax
	mov  dl,es:THISPUVAR	;number of current pu.
		;... be careful dx not overwritten below.

;make sure es set correctly...
	push ds 		;ds and es are used by stos/movs
	pop  es 		;/			instructions.

;clear the delivery variables...
	lea di,termvars
	mov cx,NUMTERMS
	xor ax,ax
	rep stosb


	mov  di,pcircin 	;leading-edge. remote i/p msgs added here.
	mov  si,di		;(we search the other way, toward pcircout)
	add  si,4		;next msg. (which is going to higher addrs)
get1:
;check limits...
	cmp  di,pcircout
	jne  get1a
get1x:	jmp   retgetmsgs		;finished>>>>>>>>>
get1a:	cmp  di,OFFSET PBOTTOM
	jb   get1b
	mov  di,OFFSET PTOP
	mov  si,OFFSET PTOP+4
	jmp  SHORT get1  ;get1c
get1b:	cmp  si,OFFSET PBOTTOM
	jb   get1c
	mov  si,OFFSET PTOP
get1c:

	mov  al,[di+3]		;msg3.
	cmp  al,0		;is it empty slot?
	jz   get3		;  remove.
	test al,80h		;is it a transform?
	jnz  get3		;  remove.
	and  al,3Fh
	  cmp  al,3Fh		;must also deliver remote i/p msgs.
	  je   get1d		;/
	cmp  al,dl  ;THISPUVAR		;check dest. pu.
	jne  get2		;wrong pu.
get1d:	mov  al,[di+2]		;msg2
	and  al,1Fh
	cmp  al,BYTE PTR destnode
	jne  get2		;wrong node.

;success, so deliver msg...
	xor  ax,ax
	mov  al,[di+2]		;get msg2
	mov  cl,5
	shr  al,cl		;terminal 1--8 (stored 0--7)
	shl  ax,1		;each pointer in termvars is 2 bytes.
	lea  bx,termvars
	add  bx,ax		;pointer to msgterm"n".
	mov  ax,[di]		;get msg0 and msg1.
	  mov cl,[di+3]		;get msg3.
	  test cl,40h		;bit6=1 if async msg.
	  jz  del0
	  inc WORD PTR [bx]		;incr count.
	  push ax		;deliver async msg to stack.
	  jmp  SHORT del10
del0:	mov  [bx],ax		;deliver sync msg.
del10:

;see if this is only msg left in buffer...
get3:	mov  ax,di
	add  ax,4
	cmp  ax,pcircout
	jne  get3p
get3q:	mov  pcircout,di	;remove it.
	jmp  SHORT get1x	;get out.
get3p:	cmp  ax,OFFSET PBOTTOM
	jne  get3s
	cmp  pcircout,OFFSET PTOP
	je   SHORT get3q

;moves all of msgs along, to fill gap...
get3s:	push si
	push di
get3a:	cmp  si,pcircout	;<<? if so, there is no next msg.
	je   get3c		;/
	cmp  si,OFFSET PBOTTOM
	jb   get3b
	mov  si,OFFSET PTOP
	jmp  SHORT get3a
get3b: ;	cmp  si,pcircout	;if so, there is no next msg.
       ;	jae  get3c
	cmp  di,OFFSET PBOTTOM
	jb   get3d
	mov  di,OFFSET PTOP
get3d:	mov  cx,4		;move next msg into empty slot.
	rep  movsb		;/
	jmp  get3a
get3c:	mov  pcircout,di	;tail msg dropped off.

;	  sub  di,4	     ;** fix occassional hanging problem.**
;	  cmp  di,pcircin    ;/	...have removed last msg,
;	  jne  get3m	     ;/      so get out.
;	  mov  pcircout,di   ;/
;	  jmp  get1x	     ;/

get3m:	pop  di
	pop  si
	jmp  get1

get2:	add  di,4
	add  si,4
	jmp  get1

 ;	ret <<<<CANNOT HAVE A RET, AS PUSHED ASYNC MSGS ON STACK

;................................................................
beepfreq PROC FAR	;used for testing purposes.bx=counter value.
	push ax
	mov al,0B6h
	out 43h,al
	mov al,bl
	out 42h,al
	mov al,bh
	out 42h,al
	in  al,61h
	or  al,03
	out 61h,al
	pop ax
	ret
beepfreq ENDP
beepoff PROC FAR
	push ax
	in  al,61h
	and al,0FCh
	out 61h,al
	pop ax
	ret
beepoff ENDP
;...................................................................

codeTERSE ENDS
;..............................................................


;###############################################################
;################### VIRTUAL PROCESSORS BELOW ##################

;######################## PROCESSOR 1 ##########################

;..........DATA SEGMENT FOR PROCESSOR 1...............
;below is the data required by TERSE, for this processor.
;we instantiate this for each processor.

dataPU1 SEGMENT  'data'

;we need somewhere to save register state when suspending this task...
;ds,es,ax,bx,cx,dx,bp,si,di,ip,cs,sp,ss,es,flags...
regs  REGSSTRUC  <>
ipnxtnode   DW	0	;used by TERSE to form a far jump address.
csthispu    DW	0    ;/ loaded by initTASKS, easy access to cs of this pu.

_fSIGFAIL    DB	0	;error flags. (I hate using a whole byte for
_fTIMEOUT    DB	0	; one flag! ...but convenient for C interface)

localref0   DB	0	;snapshot of globaltime.
localref1   DB	0	;/   (used for local timeout)
localref2   DB	0	;/
localref3   DB	0	;/

PTOP LABEL   NEAR		;start of circular buffer, for i/p msgs.
circbuf      DB  CIRCBUFSIZE DUP(0)
PBOTTOM LABEL NEAR		;one past end of circ. buffer.
;... each msg is 4 bytes, so 32 msgs will fit in buffer.
pcircin    DW	0	;ptr to top circ.buff.
pcircout   DW	0	;ptr to end circ.buff.

transform2nd DW  0	;saved transform. msg0,msg1
transform    DW	0	;/		  msg2,msg3
extranode    DB  0
destnode     DW	0
nodenumsaved DW 0
extmsgwtg    DW 0
extmsgsaved  DW 0

NUMTERMS     EQU 8
termvars     LABEL NEAR 	;msgs delivered here.
_msgterm1     DW  0		;(if an async msg, only a count
_msgterm2     DW  0		; here, msgs pushed on stack)
_msgterm3     DW	 0
_msgterm4     DW	 0
_msgterm5     DW	 0
_msgterm6     DW	 0
_msgterm7     DW	 0
_msgterm8     DW	 0

stackPU   DB  STACKSIZE DUP(?)		;stack.

;............................................................

; ***PUT STATIC APPLICATION VARIABLES FOR PU1 IN HERE***<<<<<<<<<<<<<<
beeper DW 0
togglebpr DB 0
;note: any data in small-C declared extern can be defined here,
;but must have leading underscore...
_x     DW 0

;.............................................................
dataPU1 ENDS

;................CODE FOR PROCESSOR 1...................
codePU1 SEGMENT  'code'
	ASSUME cs:codePU1,ds:dataPU1

THISPU	= 1
NUMENTRIES = 5
THISPUVAR     DB THISPU		;range is 1 - 61 (stored as 1 -- 61).
NUMENTRIESVAR DB NUMENTRIES	;number of entries in sig.table.
nodeptr	DW node0       ;1st node in table is error-handler(always node-0).
	DW node1	       ;this must be in ascending order
	DW node2	       ;of node number.
	DW node3	       ;
	DW node4
;....***KEEP GOING HERE, AS MANY NODES AS YOU'VE GOT***....

	ORG (2*64) + 2	       ;fixed offset for sig.table.
;.........***FILL IN SIGNATURE TABLE***..............
WX	   EQU 8000h		;bit15. used in nodenum.
signatures DW 0FFFFh,0,307h,322h,34Ch
nodenum    DW 0,     1, 2,   3,	  4	;*
extmsg	   DW 0,     0, 0,   0,	  0
;....................................................


;.....

;...................................................................
node0:		    ;node-0 is the error-handler

	 NORMALEXIT

;...................................................................
node1:
;Presume that this is the first node, when signature =0.

	POSTMESSAGE THISPU,2,1,SYNC,1234h ;pu1,node2,term1,sync,data=1234h.
	POSTMESSAGE THISPU,3,1,SYNC,5678h ;pu1,node3,term1,sync,data=5678h.

	.IF togglebpr==FALSE
	  mov beeper,1200	;about 830Hz ;inc beeper
	  mov togglebpr,TRUE
	.ELSE
	  mov beeper,300	;about 3.3KHz.
	  mov togglebpr,FALSE
	.ENDIF
	POSTMESSAGE 2,4,4,SYNC,beeper	;remote o/p msg.,sync/pu2/n4/t4.


	NORMALEXIT
;...................................................................
node3 PROC
;***EXAMPLE C CODING OF NODE***
;By placing this node in PROC...ENDP, code labels become invisible outside.
;... but, must use MASM v6.1, not 5.1.
;Here is our code to be replaced by C...
;	mov  ax,_msgterm1	;sync msg in terminal1.
;	POSTMESSAGE THISPU,4,1,SYNC,3344h	;sync to node4/term1.
;	NORMALEXIT

;OK, here is our C code...
INCLUDE n003.mac	;could create from GOOFEE Diagrammer.

node3 ENDP
;..................................................................
node2:
	mov  ax,_msgterm1	;sync msg in terminal1.
	POSTMESSAGE THISPU,4,3,ASYNC,8899h	;async to node4/term3.
	POSTMESSAGE THISPU,4,2,SYNC,2244h	;sync to node4/term2.
	NORMALEXIT
;...................................................................
node4:
	mov  ax,_msgterm1	;sync msg from node3.
	mov  bx,_msgterm2	;sync msg from node2
	cmp  _msgterm3,0	;count of async msgs from node2.
	je   p1n4b		;  note an async coming from pu2/n1 also.
	  mov  cx,_msgterm3
p1n4a:	  pop  ax		;get async msg off stack
	    ;....do something with it.
	  loop p1n4a
p1n4b:

	mov cx,20 ;5  ;10		;big delay.
zzz2:	mov bx,cx		; / we are posting variable beeper
	mov cx,-1		; / to pu2, and want to do it slowly
zzz1:	nop			; /
	loop zzz1		; / (beeper is posted from n1)
	mov cx,bx		; /
	loop zzz2		; /


	NORMALEXIT
;................
codePU1 ENDS


;######################## PROCESSOR 2 ##########################

;..........DATA SEGMENT FOR PROCESSOR 2...............
;we don't redefine all the variables that TERSE uses, just allocate
;a block of memory...

dataPU2 SEGMENT  'data'
;&&	DB (OFFSET enddataPU1 - OFFSET dataPU1) DUP(?)
;&&	DB enddataPU1 DUP(0)
	DB 02CFh  DUP(0)	;actually, its 2C7h.

;............................................................
; ***PUT APPLICATION VARIABLES IN HERE***<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
_z	DW 0

dataPU2 ENDS
;............................................................


;................CODE FOR PROCESSOR 2...................
codePU2 SEGMENT  'code'
	ASSUME cs:codePU2,ds:dataPU2

THISPU	= 2
NUMENTRIES = 5
	DB THISPU	;THISPUVAR	# of this pu.
	DB NUMENTRIES	;NUMENTRIESVAR. # entries in sig.table.
			   ;nodeptr
	DW pu2node0	   ;1st node in table is error-handler(always node-0).
	DW pu2node1	       ;this must be in ascending order
	DW pu2node2	       ;of node number.
	DW pu2node3	       ;
	DW pu2node4
;....***KEEP GOING HERE, AS MANY NODES AS YOU'VE GOT***....

	ORG (2*64) + 2	       ;fixed offset for sig.table.
;.........***FILL IN SIGNATURE TABLE***..............
 DW 0FFFFh,0,607h,622h,64Ch	;signatures
 DW 0,	   1, 2,   3,  WX+0	;nodenum --do NOT put 4 in last entry!
 DW 0,	   0, 0,   0,	10h	;extmsg --set bit4, last entry, as sync
				; msg from pu1/n1 to pu2/n4/t4.
				;WX -- wait is set, as cannot go any further
				;until the remote i/p arrives (bit15=1).
				;nodenum=0 means wait forever for rem.i/p.
;....................................................



;...................................................................
pu2node0:		    ;node-0 is the error-handler

	 NORMALEXIT

;...................................................................
pu2node1:
;Presume that this is the first node, when signature =0.

	POSTMESSAGE THISPU,2,1,SYNC,1234h ;pu,node2,term1,sync,data=1234h.
	POSTMESSAGE THISPU,3,1,SYNC,5678h ;pu,node3,term1,sync,data=5678h.
	POSTMESSAGE 1,4,3,ASYNC,1111h	  ;remote o/p msg.,pu1/n4/t3/async.

	NORMALEXIT
;...................................................................
pu2node3 PROC
;	mov  ax,_msgterm1	;sync msg in terminal1.
;	POSTMESSAGE THISPU,4,1,SYNC,3344h	;sync to node4/term1.
;	NORMALEXIT
;same as node3 in pu1, so....

INCLUDE n007.mac	;small-C equivalent.

pu2node3 ENDP
;..................................................................
pu2node2:
	mov  ax,_msgterm1	;sync msg in terminal1.
	POSTMESSAGE THISPU,4,3,ASYNC,8899h	;async to node4/term3.
	POSTMESSAGE THISPU,4,2,SYNC,2244h	;sync to node4/term2.
	NORMALEXIT
;...................................................................
pu2node4:
	mov  ax,_msgterm1	;sync msg from node3.
	mov  bx,_msgterm2	;sync msg from node2

	cmp  _msgterm3,0	;count of async msgs from node2.
	je   p2n4b
	  mov  cx,_msgterm3
xtract1:  pop  ax		;get async msg off stack
	    ;....do something with it.
	  loop xtract1
p2n4b:

	mov  bx,_msgterm4	;remote i/p sync msg from pu1/n1.
	call beepfreq		;uses bx as timer count.


	NORMALEXIT
;................
codePU2 ENDS


;.............................

;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	END   start

;................................................................

;Circular buffer:     |________|
;	     PTOP = > |	       |  limit
;		      |	       |
;		      |________|
;	     pcircin >|	       | pcircin is the head of the circular buffer.
;		      |	       |
;		      |________|
;	    pcircout >|	       | pcircout is the tail(one below, actually).
;		      |________|
;	 PBOTTOM = >  |	       | one past limit.
;
;VERY IMPORTANT: increasing addresses downwards.

;a message consists of 4 bytes, msg3,msg2,msg1,msg0, where
;msg3 is the msb and is at the highest address.

;ALL msgs go onto the same buffer. Remote i/p msgs add onto pcircin
;end of buffer. Local msgs add onto pcircout end of buffer.
;When getmsgs() extracts msgs from the buffer, msgs are moved up to
;fill the holes, contracting from the pcircout end.

;TIMEOUT
;*******
;"nodenum" entry in sig. table has timeouts. If bit15 set (WAIT bit)
;then bits 0--14 are timeout. Bits13--14 are granularity.
;localref0--3 is a snapshot of globaltime when timeout commences.
;BITS14,13	GLOBALTIME BYTES       TIME GRANULARITY
;00		1,0		       ...globaltime0 always=0...
;01		2,1		       1msec
;10		3,2		       1*256mSec      appr. 1/4sec
;11		3		       1*256*256mSec  appr. 1min

;The timeout value in nodenum is in bits 0--12, so max value 8191.
;so, 1msec-granularity has a timeout range of 1msec -- 8.191sec.
;
;NOTE if bits 0--12 ==0, then timeout is forever!!.





NOTES
-----

CODEVIEW 4.01 (DOS version)
I have a problem with Codeview, as it hooks int8, before I do.
As I have speeded up the timer 55 times, CV behaves a little oddly
-- scrolls windows like crazy!
Recommend you mess around with this code on a PC that doesn't have
your company accounts!
Then, CV calls the application's int8 at a very slow rate....
I guess the idea behind this is you can observe any effect visually,
but I wish Microsoft would document how CV handles int8.

MASM
The program is written for MASM v6.1. I have TASM v3.1, which isn't
suitable as it doesn't have the nifty .IF/.ELSEIF/.ELSE/.ENDIF
directives. I don't know about the latest TASM -- maybe someone can
enlighten me.


;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


COPYRIGHT & LIABILITY NOTICE
****************************
TERSE may be freely used, as long as the name "TERSE" is kept,
with suitable pre- & postfix characters, and Barry kauler is
acknowledged as the original developer.  Any publications about
TERSE, or a TERSE derivative, must reference Barry Kauler's
Web site, or one of his publications.

Barry Kauler accepts no liability for how this product performs.

(c) Copyright Barry Kauler 1995
email:	B.KAULER@COWAN.EDU.AU
WWW:	 http://scorpion.cowan.edu.au/science/terse/terse.htm


POSTTRANSFORM MACRO signature,nodet,timet,timewindow,xtranode
     LOCAL trn1
     IF ((nodet LT 0) OR (nodet GT 31)) 	     ;assemble-time test.
       EXITM					     ;/
     ELSEIF ((timewindow LT 0) OR (timewindow GT 31)) ;/
       EXITM					      ;/
     ELSEIF ((xtranode LT 0) OR (xtranode GT 31))     ;/
       EXITM					      ;/
     ELSEIF (signature GT 07FFFh)		      ;/
       EXITM					      ;/
     ELSE
	mov  bx,pcircout
	cmp  bx,OFFSET PBOTTOM	     ;tail-end of circ.buffer.
	jne  trn1
	mov  bx,OFFSET PTOP		      ;(note that PTOP is lowest address).
trn1:
	mov  [bx],BYTE PTR OFFSET ((timewindow*32) + (xtranode)) ;msg0
	inc  bx
	mov  [bx],OFFSET timet			;msg1
	inc  bx
      IF (signature AND 4000h)		;test if bit14 set.
 mov  [bx],WORD PTR OFFSET ((signature XOR 0010010000010001b) XOR (nodet)) + 8000h
	inc  bx		 ;...msg2,msg3. bit15 must be set.
      ELSE
	mov  [bx],WORD PTR OFFSET (signature XOR (nodet)) + 8000h
	inc  bx 	 ;...msg2,msg3.  bit15 must be set.
      ENDIF
	inc  bx
	mov  pcircout,bx		;new tail-end.
     ENDIF
ENDM
