/* GOOFEE Diagrammer introductory example. */

/* DEMO1.GOO is a part of a solution for a baby             */
/* breathing monitor.                                       */
/* Two full solutions, one requiring an RTOS, with          */
/* multiple threads, and the other not needing any          */
/* underlying kernel, are to be published in a journal.     */

/* This file, DEMO1.C, is the file with main().             */
/* It also has some interthread functions.                  */
/* Some of the #defines below are system dependent.         */
/* ...don't be distracted by them, as they are "noise" and  */
/* not directly relevant to this intro.                     */
/* The complete project consists of this file, DEMO1.H,     */
/* SEQUENCER.C, and ISR.C.                                  */

/* Before you do what I suggest below, note that the        */
/* two source code files you will generate are ISR.C        */
/* for ISR thread, and SEQUENCER.C for SEQUENCER thread.    */
/* normally, you would compile the three files separately   */
/* and link them together -- note that you can even run a   */
/* batch file or make file from a node -- refer to the      */
/* node dialog box by right-clicking on a node.             */
/* now do something interesting...                          */

/* To see some generated code, get out of this file,        */
/* left-click on "ISR" or "SEQUENCER", then select          */
/* "File/generate 1-thread code", then double-click on "ISR"*/
/* or "SEQUENCER" to view the generated code.               */

/* you can run these files independently through your C compiler. */

/* You can keep expanding this diagram, right down to       */
/* 100% code generation.  Alternatively, you can insert     */
/* code blocks -- a file with name "NxxxDzzz.eee", or       */
/* "NxxxDzzz.Ryy", where zzz = diagram number,              */
/* xxx = node number, yy = ring number, eee = extension     */
/* (default is .C) -- is auto-generated, simply by          */
/* double-clicking on the node, within the appropriate ring.*/

/*........................................*/
#include "demo1.h"


/*this array hold flags for the interthread messages...*/
int  msgs[16];

void main(void) {
  SEQUENCER();		/*never comes back.*/
}

/* these are the functions for posting and receiving*/
/*interthread messages...*/
/*(prototypes in DEMO1.H)*/

/*post a message. i is the destination node number...*/
int postmsg(int i) {
  int ii,iii,xx;
  ii = i/16;
  iii = i%16;
  do {
    xx = msgs[ii] >> iii; /*move the flag to right side.*/
    xx = xx & 01;   /*mask off all other flags.*/
  }  while (xx);		/*in case previous msg not consumed.*/
      xx = xx | 01;		/*set the bit.*/
      xx = xx << iii;  /*move it back to correct bit-position.*/
  INTOFF; /*only need to do this if following statement doesn't*/
		/*compile to a single instruction...*/
      msgs[ii] = msgs[ii] | xx;	  /*set the flag.*/
    INTON;
  return 1;  
}


/*receives an interthread message.*/
/*only returns when message has arrived.*/
/*in future, this could have a timeout, to implement the*/
/*sync wire with reply & timeout.*/
int getmsg(int i) {
  int ii,iii,xx,fmsg;
  ii =i + i/16; 	/*index into array.*/
  iii = i%16;	/*remainder.*/
  do {
    xx = msgs[ii] >> iii;		/*get flag into bit0.*/
    fmsg = xx & 01;		/*extract bit0 only.*/
  }  while (!fmsg);
    xx = fmsg << iii;	/*move logic-1 into correct bit position.*/
  INTOFF; /*some cpus and optimising compilers will not need to turn off*/
	/*interrupts, as following will be atomic, ie, one instruction...*/
    msgs[ii] = msgs[ii] ^ xx;		/*clr the bit (exclusive or)*/
  INTON;
  return fmsg;
}


/*returns immediately, with flag if message has arrived.*/
/*the code generator assigns this flag to fASYNC.*/
int ismsg(int i) {
  int ii,iii,xx,fmsg;
  ii =i + i/16; 	/*index into array.*/
  iii = i%16;	/*remainder.*/
    xx = msgs[ii] >> iii;		/*get flag into bit0.*/
    fmsg = xx & 01;		/*extract bit0 only.*/
    xx = fmsg << iii;	/*move logic-1 into correct bit position.*/
  if (fmsg) {
    INTOFF; /*some cpus and optimising compilers will not need to turn off*/
	    /*interrupts, as following will be atomic, ie, one instruction...*/
      msgs[ii] = msgs[ii] ^ xx;		/*clr the bit (exclusive or)*/
    INTON;
  }
  return fmsg;
}

/* NOTES */
/* you will notice in the GOOFEE Diagrammer that the   */
/* data types "uchar" and "uint" are used. These are   */
/* short for "unsigned char" and "unsigned int", but   */
/* you must have typedef's in the header file.         */
/* you will see them in DEMO1.H.                       */


